# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_wrds/01_crspm.ipynb.

# %% ../../nbs/01_wrds/01_crspm.ipynb 3
from __future__ import annotations
from pathlib import Path
from typing import List
import os

import pandas as pd
import numpy as np

import pandasmore as pdm
from . import wrds_api
from .. import RESOURCES

# %% auto 0
__all__ = ['raw_metadata', 'raw_metadata_extra', 'default_raw_vars', 'parse_varlist', 'delist_adj_ret', 'download', 'clean']

# %% ../../nbs/01_wrds/01_crspm.ipynb 4
def raw_metadata(rawfile: str|Path=RESOURCES/'crspm_variable_descriptions.csv', # location of the raw variable labels file
             ) -> pd.DataFrame:
    "Loads raw variable labels file, cleans it and returns it as a pd.DataFrame"

    df = pd.read_csv(rawfile)
    df['output_of'] = 'wrds.crspm.clean'

    df['Variable Label'] = df.apply(lambda row: row['Description'].replace(row['Variable Name'].strip()+' -- ', ''), axis=1)
    df['Variable Label'] = df.apply(lambda row: row['Variable Label'].replace( '(' + row['Variable Name'].strip() + ')', ''), axis=1)
    df['Variable Name'] = df['Variable Name'].str.strip().str.lower()
    df = df[['Variable Name', 'Variable Label','output_of', 'Type']].copy()
    df.columns = ['name','label','output_of','type']
    return df

# %% ../../nbs/01_wrds/01_crspm.ipynb 7
def raw_metadata_extra(wrds_username: str=None
             ) -> pd.DataFrame:
    "Collects metadata from WRDS `crsp.msf` and `crsp.msenames` tables and merges it with `variable_labels`."

    if wrds_username is None:
        wrds_username = os.getenv('WRDS_USERNAME')
        if wrds_username is None: wrds_username = input("Enter your WRDS username: ") 

    try:
        db = wrds_api.Connection(wrds_username = wrds_username)
        msf = db.describe_table('crsp','msf')
        msf_rows = db.get_row_count('crsp','msf')
        mse = db.describe_table('crsp','msenames')
        mse_rows = db.get_row_count('crsp','msenames')
    finally:
        db.close()
        
    msf_meta = msf[['name','type']].copy()
    msf_meta['nr_rows'] = msf_rows
    msf_meta['wrds_library'] = 'crsp'
    msf_meta['wrds_table'] = 'msf'

    mse_meta = mse[['name','type']].copy()
    mse_meta['nr_rows'] = mse_rows
    mse_meta['wrds_library'] = 'crsp'
    mse_meta['wrds_table'] = 'msenames'

    crsp_meta = (pd.concat([msf_meta, mse_meta],axis=0, ignore_index=True)
                .merge(raw_metadata()[['name','label']], how='left', on='name'))
    
    crsp_meta['output_of'] = 'wrds.crspm.download()'
    crsp_meta = pdm.order_columns(crsp_meta,these_first=['name','label','output_of'])
    for v in list(crsp_meta.columns):
        crsp_meta[v] = crsp_meta[v].astype('string')
    
    return crsp_meta

# %% ../../nbs/01_wrds/01_crspm.ipynb 9
def default_raw_vars():
    """Default variables used in `download` if none are specified. Takes about 2 min to download."""
    
    return ['permno','permco','date',
            'ret', 'retx', 'shrout', 'prc', 
            'shrcd', 'exchcd','siccd','ticker','cusip','ncusip']            

# %% ../../nbs/01_wrds/01_crspm.ipynb 11
def parse_varlist(vars: List[str]=None,
                  wrds_username: str=None
                  ) -> str:
    """Figure out which `vars` come from the `crsp.msf` table and which come from the `crsp.msenames` table and add a. and b. prefixes"""

    if wrds_username is None:
        wrds_username = os.getenv('WRDS_USERNAME')
        if wrds_username is None: wrds_username = input("Enter your WRDS username: ") 

    if vars is None: vars = default_raw_vars()
    vars = ['permno','permco','date','exchcd'] + [x for x in vars if x not in ['permno','permco','date','exchcd']]

    try:
        db = wrds_api.Connection(wrds_username = wrds_username)
        all_msf_vars = list(db.describe_table('crsp','msf').name)
        all_mse_vars = list(db.describe_table('crsp','msenames').name)
        my_msf_vars = [f'a.{x}' for x in vars if x in all_msf_vars]
        my_mse_vars = [f'b.{x}' for x in vars if (x in all_mse_vars) and (x not in all_msf_vars)]
        varlist_string = ','.join(my_msf_vars + my_mse_vars)
    finally:
        db.close()
        
    return varlist_string

# %% ../../nbs/01_wrds/01_crspm.ipynb 12
def delist_adj_ret(df: pd.DataFrame, # Requires `ret`,`exchcd`, ` `dlret`, and `dlstcd` variables
                       adj_ret_var: str
                       ) -> pd.DataFrame:
    """Adjust for delisting returns using Shumway and Warther (1999) and Johnson and Zhao (2007)"""

    df['npdelist'] = (df['dlstcd']==500) | df['dlstcd'].between(520,584)
    df['dlret'] = np.where(df.dlret.isna() & df.npdelist & df.exchcd.isin([1,2]), -0.35, df.dlret)
    df['dlret'] = np.where(df.dlret.isna() & df.npdelist & df.exchcd.isin([3]), -0.55, df.dlret)
    df['dlret'] = np.where(df.dlret.notna() & df.dlret < -1, -1, df.dlret)
    df['dlret'] = df.dlret.fillna(0)

    df[adj_ret_var] = (1 + df.ret) * (1 + df.dlret) - 1
    df[adj_ret_var] = np.where(df[adj_ret_var].isna() & (df.dlret!=0), df.dlret, df[adj_ret_var])
    df = df.drop('npdelist', axis=1) 
    return df

# %% ../../nbs/01_wrds/01_crspm.ipynb 13
def download(vars: List[str]=None, # If None, downloads `default_raw_vars`; `permno`, `permco`, `date`, and 'exchcd' are added by default
             obs_limit: int=None,  #Number of rows to download. If None, full dataset will be downloaded             
             wrds_username: str=None,       #If None, looks for WRDS_USERNAME with `os.getenv`, then prompts you if needed
             start_date: str="01/01/1900",  # Start date in MM/DD/YYYY format
             end_date: str=None,            # End date in MM/DD/YYYY format; if None, defaults to current date  
             add_delist_adj_ret: bool=True, # Whether to calculate delisting-adjusted returns 
             adj_ret_var: str='ret_adj'     # What to call the returns adjusted for delisting bias
             ) -> pd.DataFrame:
    """Downloads `vars` from `start_date` to `end_date` from WRDS crsp.msf and crsp.msenames libraries. 
        Creates `ret_adj` for delisting based on Shumway and Warther (1999) and Johnson and Zhao (2007)"""

    varlist_string = parse_varlist(vars, wrds_username)
    limit_clause = f"LIMIT {obs_limit}" if obs_limit is not None else ""
    sql_string = f"""SELECT {varlist_string},  c.dlstcd, c.dlret 
                        FROM crsp.msf AS a 
                        LEFT JOIN crsp.msenames AS b
                            ON a.permno=b.permno AND b.namedt<=a.date AND a.date<=b.nameendt                     
                        LEFT JOIN crsp.msedelist as c
                            ON a.permno=c.permno AND date_trunc('month', a.date) = date_trunc('month', c.dlstdt)                            
                            WHERE a.date BETWEEN '{start_date}' AND COALESCE(%(end)s, CURRENT_DATE) 
                    {limit_clause}
                """
    df = wrds_api.download(sql_string, wrds_username=wrds_username, params={'end':end_date})
    if add_delist_adj_ret: df = delist_adj_ret(df, adj_ret_var)
    else: df = df.drop(['dlret','dlstcd'], axis=1)
    return df 

# %% ../../nbs/01_wrds/01_crspm.ipynb 16
def clean(df: pd.DataFrame=None,        # If None, downloads `vars` using `download` function; else, must contain `permno` and `date` columns
          vars: List[str]=None,         # If None, downloads `default_raw_vars`
          obs_limit: int=None, #Number of rows to download. If None, full dataset will be downloaded
          wrds_username: str=None,      # If None, looks for WRDS_USERNAME with `os.getenv`, then prompts you if needed
          start_date: str="01/01/1900", # Start date in MM/DD/YYYY format
          end_date: str=None,           # End date. Default is current date          
          clean_kwargs: dict={},        # Params to pass to `pdm.setup_panel` other than `panel_ids`, `time_var`, and `freq`
          ) -> pd.DataFrame:
    """Applies `pandasmore.setup_panel` to `df`. If `df` is None, downloads `vars` using `download` function."""

    if df is None: df = download(vars=vars, obs_limit=obs_limit, wrds_username=wrds_username, start_date=start_date, end_date=end_date)
    df = pdm.setup_panel(df, panel_ids='permno', time_var='date', freq='M', **clean_kwargs)
    return df 



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/cli.ipynb.

# %% auto 0
__all__ = ['search']

# %% ../nbs/cli.ipynb 2
from fastcore.script import call_parse
from nbdev import nbdev_export
from nbdev.quarto import refresh_quarto_yml, nbdev_readme

# %% ../nbs/cli.ipynb 3
@call_parse
def search(label):
    "Fuzzy search of variable labels for a given string"
    
    from finsets import search
    return search(label)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_core.ipynb.

# %% ../nbs/00_core.ipynb 4
from __future__ import annotations
from typing import Literal
from importlib import import_module
from inspect import signature

import pandas as pd
from  thefuzz import process, fuzz

from fastcore.script import call_parse

# %% auto 0
__all__ = ['features_metadata', 'raw_metadata', 'all_metadata', 'search']

# %% ../nbs/00_core.ipynb 5
def features_metadata(submodules: list=['wrds', 'papers'] # list of submodules to collect metadata from
                      ) -> pd.DataFrame:
    "Go through `submodules` of `finsets` and collect metadata from all functions that have `return_metadata` parameter"
    
    df = pd.DataFrame(columns=['name','label','output_of','inputs','inputs_generated_by'])
    for name in submodules:
        module = import_module(f'finsets.{name}')
        for sub in dir(module):
            if sub.startswith('_'): continue
            submodule = import_module(f'finsets.{name}.{sub}')
            for func_name in submodule.__all__:
                func = getattr(submodule, func_name)
                if callable(func):
                    try: 
                        params = signature(func).parameters
                    except:
                        continue
                    if 'return_metadata' in params: 
                        meta = func(return_metadata=True)
                        for var_name in meta['outputs']:
                            for input_name in meta['inputs']:    
                                new_meta = pd.DataFrame({'name':var_name, 
                                                'label':meta['labels'][var_name], 
                                                'output_of':f'{name}.{sub}.{func_name}', 
                                                'inputs':','.join(meta['inputs'][input_name]),
                                                'inputs_generated_by':input_name}, index=[0])
                                df = pd.concat([df,new_meta],ignore_index=True)
    return df

# %% ../nbs/00_core.ipynb 7
def raw_metadata(submodules=['wrds', 'papers'] # list of submodules to collect metadata from
                ) -> pd.DataFrame:
    "Go through `submodules` of `finsets` and collect metadata from `raw_metadata` functions (if present)"

    df = pd.DataFrame(columns=['name','label','output_of','type'])
    for name in submodules:
        module = import_module(f'finsets.{name}')
        for sub in dir(module):
            if sub.startswith('_'): continue
            submodule = import_module(f'finsets.{name}.{sub}')
            if 'raw_metadata' in submodule.__all__:
                df = pd.concat([df,submodule.raw_metadata()],ignore_index=True)
    return df

# %% ../nbs/00_core.ipynb 9
def all_metadata(submodules=['wrds', 'papers'] # list of submodules to collect metadata from
                ) -> pd.DataFrame:
    "Collects `raw_metadata` and `features_metadata` from `submodules` of `finsets`"

    return pd.concat([features_metadata(submodules), raw_metadata(submodules)], ignore_index=True)

# %% ../nbs/00_core.ipynb 11
@call_parse
def search(query: str,              # What to search for 
           meta: str='all',   #"all", "features", or "raw"; specifies the function that fetches the metadata you want to search through
           field: str='label', # Which column in the metadata table you want to search through. Use "name" to search variable names.  
           limit: int=10,           # How many results to display                      
           ) -> pd.DataFrame:
    "Search for `query` in metadata returned by f`{meta_func}_metadata`; return `limit` number of results"

    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', 100)

    try:   
        import finsets
        metadata = getattr(finsets, f'{meta}_metadata')()
        results = process.extractBests(query, metadata[field], 
                                        scorer = fuzz.token_sort_ratio,
                                        limit=limit)
        rows = [x[2] for x in results]
        scores = [x[1] for x in results]
        df = metadata.iloc[rows]
        df.index = scores
        df.index.name="SCORE"
        df.columns = df.columns.str.upper()
    finally:
        pd.reset_option('display.max_columns')
        pd.reset_option('display.width')        
    return df
       



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/00_fred/00_fred.ipynb.

# %% auto 0
__all__ = ['download', 'clean', 'get_series_info', 'search', 'common_raw_vars']

# %% ../../nbs/00_fred/00_fred.ipynb 4
import os, time

import pandas as pd
from .fred_api import Fred

import pandasmore as pdm

# %% ../../nbs/00_fred/00_fred.ipynb 5
def download(series: str=None, # FRED series name
               api_key: str=None # FRED api key. If None, will use os.getenv("FRED_API_KEY")
               ) -> pd.DataFrame: 
    """Retrieves series from FRED and returns it as a pd.DataFrame."""

    if api_key is None: api_key = os.getenv("FRED_API_KEY")
    api = Fred(api_key=api_key)
    return api.get_series(series).to_frame(name=series)

# %% ../../nbs/00_fred/00_fred.ipynb 7
def clean(series: str=None, # FRED series name
               label: str=None, # Name you want to give to the series in the output DataFrame. If None, use lowercase of `series`
               api_key: str=None # FRED api key. If None, will use os.getenv("FRED_API_KEY")
               ) -> pd.DataFrame: 
    """Retrieves series from FRED, cleans the date and sets it as index"""

    if api_key is None: api_key = os.getenv("FRED_API_KEY")
    api = Fred(api_key=api_key)
    freq = api.get_series_info('GDP').frequency_short 
    time.sleep(1)
    if label is None: label = str.lower(series)
    df = download(series).rename(columns={series:label}).reset_index().rename({'index':'date'},axis=1)
    df = pdm.setup_tseries(df, freq=freq).drop('date', axis=1)
    return df 

# %% ../../nbs/00_fred/00_fred.ipynb 9
def get_series_info(series: str=None, # FRED series name
                    api_key: str=None # FRED api key. If None, will use os.getenv("FRED_API_KEY")
                    ) -> pd.Series:
    """Get metadata from FRED for given `series` from FRED"""

    if api_key is None: api_key = os.getenv("FRED_API_KEY")
    api = Fred(api_key=api_key)
    return api.get_series_info(series)

# %% ../../nbs/00_fred/00_fred.ipynb 11
def search(search_text: str=None, # What to search for
              order_by: str='popularity', # How to order search results; try `search_rank` if you don't find what you were looking for
              nr_results: int=10, # How many results to output
              api_key: str=None # FRED api key. If None, will use os.getenv("FRED_API_KEY")
              ) -> pd.DataFrame:
    """Search FRED for a given `search_text`, sort by popularity and return only the first `nr_results`"""

    if api_key is None: api_key = os.getenv("FRED_API_KEY")
    api = Fred(api_key=api_key)  
    return api.search(search_text, order_by=order_by)\
              .pipe(pdm.order_columns, ['title', 'popularity','frequency_short', 'observation_start', 'observation_end'])\
              .iloc[:nr_results]
              

# %% ../../nbs/00_fred/00_fred.ipynb 14
def common_raw_vars() -> pd.DataFrame:
    """List of FRED series commonly used in finance research"""

    vars = [
    ['TB3MS','yield_3mt','3-month treasury bill secondary market rate', 'M', 'NSA',1934],
    ['DTB3','yield_3mt','3-month treasury bill secondary market rate', 'D', 'NSA',1954],
    ['GS10','yield_10yt','10-year treasury constant maturity market yield', 'M', 'NSA',1953],
    ['DGS10','yield_10yt', '10-year treasury constant maturity market yield', 'D', 'NSA',1962],
    ['GS1','yield_1yt', '1-year treasury constant maturity market yield', 'M', 'NSA', 1953],
    ['DGS1','yield_1yt', '1-year treasury constant maturity market yield', 'D', 'NSA', 1962],
    ['AAA','yield_aaa', 'Moodys Aaa seasoned corporate bond yield', 'M', 'NSA', 1919], 
    ['BAA','yield_baa', 'Moodys Baa seasoned corporate bond yield', 'M', 'NSA', 1919], 
    ['DAAA','yield_aaa', 'Moodys Aaa seasoned corporate bond yield', 'D', 'NSA', 1983], 
    ['DBAA','yield_baa', 'Moodys Baa seasoned corporate bond yield', 'D', 'NSA', 1986],     
    ['FEDFUNDS','yield_fedf', 'federal funds effective rate', 'M', 'NSA', 1954], 
    ['DFF','yield_fedf', 'federal funds effective rate', 'D', 'NSA', 1954], 
    ['CPIAUCSL','cpi', 'cpi all urban consumers', 'M', 'SA', 1947],  
    ['CPIAUCNS','cpi', 'cpi all urban consumers', 'M', 'NSA', 1913],  
    ['INDPRO','indprod', 'industrial production', 'M', 'SA', 1919],
    ['IPB50001SQ','indprod', 'industrial production', 'Q', 'SA', 1919],
    ['UNRATE','unemp', 'unemployment rate', 'M', 'SA', 1948], 
    ['GDP','ngdp', 'nominal gdp','Q','SA',1947],
    ['GDPC1','rgdp', 'real gdp','Q','SA',1947],
    ['GNP','ngnp', 'nominal gnp','Q','SA',1947],
    ['GNPC96','rgnp', 'real gnp','Q','SA',1947],
    ['GDPPOT','pot_rgdp','real potential gdp','Q','SA',1949],
    ['USREC','rec_dum', 'nber recession indicator', 'M', 'NSA', 1854], 
    ['RECPROUSM156N','rec_prob', 'nber real time recession probability', 'M', 'NSA', 1967],
    ['CFNAI','cfnai','chicago fed national activity index','M','NSA',1967],
    ['UMCSENT','sent_mich','U Michigan consumer confidence index','M','NSA',1977],
    ['MICH','exp_inflation','U Michigan inflation expectation','M','NSA',1978],
    ['USEPUINDXM','pu_bbd','overall EPU index from Baker,Bloom,Davis','M', 'NSA',1985],
    ['USEPUNEWSINDXM','punews_bbd','news comp of EPU index from BBD','M', 'NSA',1985],
    ['USEPUINDXD','pu_bbd','D EPU from BBD','D','NSA',1985],
    ['VIXCLS','vix','VIX index on snp100 from CBOE 1month','D','NSA',1990],
    ['VXOCLS','vxo','VXO index on snp500 from CBOE','D','NSA',1986]
    ]
    return pd.DataFrame(data = vars, columns = ['id','label','desc','freq','season_adj','starts']) 




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/02_papers/hassan_etal_2019.ipynb.

# %% ../../nbs/02_papers/hassan_etal_2019.ipynb 4
from __future__ import annotations
import pandas as pd

import pandasmore as pdm
from .. import wrds

# %% auto 0
__all__ = ['source_url', 'variables', 'download', 'clean']

# %% ../../nbs/02_papers/hassan_etal_2019.ipynb 5
def source_url():
    """URL where data can be downloaded from. All vintages I downloaded before are included in ascending timeline."""
    
    return pd.Series({'07_08_2023': "https://www.dropbox.com/s/m7o9oycj49rpl9d/firmquarter_2022q1.dta?raw=1"})

# %% ../../nbs/02_papers/hassan_etal_2019.ipynb 10
def variables():
    """Names of key variables in the dataset. 
    `company_name`,`hqcountrycode`,`isin`,`cusip`,`ticker` are also available but are omitted here to speed things up and save memory."""
    
    return ['gvkey','date','date_earningscall',
            'PRisk','NPRisk','Risk',
            'PSentiment','NPSentiment','Sentiment',
            'PRiskT_economic','PRiskT_environment','PRiskT_trade','PRiskT_institutions','PRiskT_health','PRiskT_security','PRiskT_tax','PRiskT_technology']

# %% ../../nbs/02_papers/hassan_etal_2019.ipynb 12
def download(url: str=source_url()[-1], # URL to the Stata (.dta) version of the dataset
            vars: list=variables(), # Which variables to download
            ) -> pd.DataFrame:
    """Download raw data from `url`"""
    
    return pd.read_stata(url, columns=vars)

# %% ../../nbs/02_papers/hassan_etal_2019.ipynb 16
def clean(df: pd.DataFrame=None, # If None, will download using `download_raw`
          gvkey_permno_link: bool|pd.DataFrame=True, # Whether to download permno or not. If DataFrame, must contain `permno`, `gvkey`, and `Qdate`
          how: str='inner' # How to merge permno into `df` if `gvkey_permno_link` is not False
          ) -> pd.DataFrame:
    """Converts `gvkey` to string and applies `pandasmore.setup_panel`. Adds `permno` if `gvkey_permno_link` is not False."""

    if df is None: df = download()
    else: df = df.copy()
    df['gvkey'] = df['gvkey'].astype('string')
    df = pdm.setup_panel(df, panel_ids='gvkey', time_var='date', freq='Q',
                        panel_ids_toint=False,
                        drop_index_duplicates=True, duplicates_which_keep='last')
    if not gvkey_permno_link: return df
    else:    
      if gvkey_permno_link is True: gvkey_permno_link = wrds.linking.gvkey_permno_q()
      df = df.reset_index().merge(gvkey_permno_link, how=how, on=['gvkey','Qdate'])
      return pdm.setup_panel(df, panel_ids='permno', dates_processed=True, freq='Q',
                              drop_index_duplicates=True, duplicates_which_keep='last')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/00_fred/01_fred_api.ipynb.

# %% auto 0
__all__ = ['urlopen', 'quote_plus', 'urlencode', 'HTTPError', 'Fred']

# %% ../../nbs/00_fred/01_fred_api.ipynb 3
import os
import xml.etree.ElementTree as ET
import urllib.request as url_request
import urllib.parse as url_parse
import urllib.error as url_error

import pandas as pd

# %% ../../nbs/00_fred/01_fred_api.ipynb 4
urlopen = url_request.urlopen
quote_plus = url_parse.quote_plus
urlencode = url_parse.urlencode
HTTPError = url_error.HTTPError

# %% ../../nbs/00_fred/01_fred_api.ipynb 5
class Fred:
    earliest_realtime_start = '1776-07-04'
    latest_realtime_end = '9999-12-31'
    nan_char = '.'
    max_results_per_request = 1000
    root_url = 'https://api.stlouisfed.org/fred'

    def __init__(self,
                 api_key=None,
                 api_key_file=None):
        """
        Initialize the Fred class that provides useful functions to query the Fred dataset. You need to specify a valid
        API key in one of 3 ways: pass the string via api_key, or set api_key_file to a file with the api key in the
        first line, or set the environment variable `FRED_API_KEY` to the value of your api key. You can sign up for a
        free api key on the Fred website at http://research.stlouisfed.org/fred2/
        """
        self.api_key = None
        if api_key is not None:
            self.api_key = api_key
        elif api_key_file is not None:
            f = open(api_key_file, 'r')
            self.api_key = f.readline().strip()
            f.close()
        else:
            self.api_key = os.environ.get('FRED_API_KEY')

        if self.api_key is None:
            import textwrap
            raise ValueError(textwrap.dedent("""\
                    You need to set a valid API key. You can set it in 3 ways:
                    pass the string with api_key, or set api_key_file to a
                    file with the api key in the first line, or set the
                    environment variable 'FRED_API_KEY' to the value of your
                    api key. You can sign up for a free api key on the Fred
                    website at http://research.stlouisfed.org/fred2/"""))

    def __fetch_data(self, url):
        """
        helper function for fetching data given a request URL
        """
        url += '&api_key=' + self.api_key
        try:
            response = urlopen(url)
            root = ET.fromstring(response.read())
        except HTTPError as exc:
            root = ET.fromstring(exc.read())
            raise ValueError(root.get('message'))
        return root

    def _parse(self, date_str, format='%Y-%m-%d'):
        """
        helper function for parsing FRED date string into datetime
        """
        rv = pd.to_datetime(date_str, format=format, errors='coerce')
        if hasattr(rv, 'to_pydatetime'):
            rv = rv.to_pydatetime()
        return rv

    def get_series_info(self, series_id):
        """
        Get information about a series such as its title, frequency, observation start/end dates, units, notes, etc.

        Parameters
        ----------
        series_id : str
            Fred series id such as 'CPIAUCSL'

        Returns
        -------
        info : Series
            a pandas Series containing information about the Fred series
        """
        url = "%s/series?series_id=%s" % (self.root_url, series_id)
        root = self.__fetch_data(url)
        if root is None or not len(root):
            raise ValueError('No info exists for series id: ' + series_id)
        info = pd.Series(list(root)[0].attrib)
        return info

    def get_series(self, series_id, observation_start=None, observation_end=None, **kwargs):
        """
        Get data for a Fred series id. This fetches the latest known data, and is equivalent to get_series_latest_release()

        Parameters
        ----------
        series_id : str
            Fred series id such as 'CPIAUCSL'
        observation_start : datetime or datetime-like str such as '7/1/2014', optional
            earliest observation date
        observation_end : datetime or datetime-like str such as '7/1/2014', optional
            latest observation date
        kwargs : additional parameters
            Any additional parameters supported by FRED. You can see https://api.stlouisfed.org/docs/fred/series_observations.html for the full list

        Returns
        -------
        data : Series
            a Series where each index is the observation date and the value is the data for the Fred series
        """
        url = "%s/series/observations?series_id=%s" % (self.root_url, series_id)
        if observation_start is not None:
            observation_start = pd.to_datetime(observation_start,
                                               errors='raise')
            url += '&observation_start=' + observation_start.strftime('%Y-%m-%d')
        if observation_end is not None:
            observation_end = pd.to_datetime(observation_end, errors='raise')
            url += '&observation_end=' + observation_end.strftime('%Y-%m-%d')
        if kwargs.keys():
            url += '&' + urlencode(kwargs)
        root = self.__fetch_data(url)
        if root is None:
            raise ValueError('No data exists for series id: ' + series_id)
        data = {}
        for child in root:
            val = child.get('value')
            if val == self.nan_char:
                val = float('NaN')
            else:
                val = float(val)
            data[self._parse(child.get('date'))] = val
        return pd.Series(data)

    def get_series_latest_release(self, series_id):
        """
        Get data for a Fred series id. This fetches the latest known data, and is equivalent to get_series()

        Parameters
        ----------
        series_id : str
            Fred series id such as 'CPIAUCSL'

        Returns
        -------
        info : Series
            a Series where each index is the observation date and the value is the data for the Fred series
        """
        return self.get_series(series_id)

    def get_series_first_release(self, series_id):
        """
        Get first-release data for a Fred series id. This ignores any revision to the data series. For instance,
        The US GDP for Q1 2014 was first released to be 17149.6, and then later revised to 17101.3, and 17016.0.
        This will ignore revisions after the first release.

        Parameters
        ----------
        series_id : str
            Fred series id such as 'GDP'

        Returns
        -------
        data : Series
            a Series where each index is the observation date and the value is the data for the Fred series
        """
        df = self.get_series_all_releases(series_id)
        first_release = df.groupby('date').head(1)
        data = first_release.set_index('date')['value']
        return data

    def get_series_as_of_date(self, series_id, as_of_date):
        """
        Get latest data for a Fred series id as known on a particular date. This includes any revision to the data series
        before or on as_of_date, but ignores any revision on dates after as_of_date.

        Parameters
        ----------
        series_id : str
            Fred series id such as 'GDP'
        as_of_date : datetime, or datetime-like str such as '10/25/2014'
            Include data revisions on or before this date, and ignore revisions afterwards

        Returns
        -------
        data : Series
            a Series where each index is the observation date and the value is the data for the Fred series
        """
        as_of_date = pd.to_datetime(as_of_date)
        df = self.get_series_all_releases(series_id)
        data = df[df['realtime_start'] <= as_of_date]
        return data

    def get_series_all_releases(self, series_id, realtime_start=None, realtime_end=None):
        """
        Get all data for a Fred series id including first releases and all revisions. This returns a DataFrame
        with three columns: 'date', 'realtime_start', and 'value'. For instance, the US GDP for Q4 2013 was first released
        to be 17102.5 on 2014-01-30, and then revised to 17080.7 on 2014-02-28, and then revised to 17089.6 on
        2014-03-27. You will therefore get three rows with the same 'date' (observation date) of 2013-10-01 but three
        different 'realtime_start' of 2014-01-30, 2014-02-28, and 2014-03-27 with corresponding 'value' of 17102.5, 17080.7
        and 17089.6

        Parameters
        ----------
        series_id : str
            Fred series id such as 'GDP'
        realtime_start : str, optional
            specifies the realtime_start value used in the query, defaults to the earliest possible start date allowed by Fred
        realtime_end : str, optional
            specifies the realtime_end value used in the query, defaults to the latest possible end date allowed by Fred

        Returns
        -------
        data : DataFrame
            a DataFrame with columns 'date', 'realtime_start' and 'value' where 'date' is the observation period and 'realtime_start'
            is when the corresponding value (either first release or revision) is reported.
        """
        if realtime_start is None:
            realtime_start = self.earliest_realtime_start
        if realtime_end is None:
            realtime_end = self.latest_realtime_end
        url = "%s/series/observations?series_id=%s&realtime_start=%s&realtime_end=%s" % (self.root_url,
                                                                                         series_id,
                                                                                         realtime_start,
                                                                                         realtime_end)
        root = self.__fetch_data(url)
        if root is None:
            raise ValueError('No data exists for series id: ' + series_id)
        data = {}
        i = 0
        for child in root:
            val = child.get('value')
            if val == self.nan_char:
                val = float('NaN')
            else:
                val = float(val)
            realtime_start = self._parse(child.get('realtime_start'))
            # realtime_end = self._parse(child.get('realtime_end'))
            date = self._parse(child.get('date'))

            data[i] = {'realtime_start': realtime_start,
                       # 'realtime_end': realtime_end,
                       'date': date,
                       'value': val}
            i += 1
        data = pd.DataFrame(data).T
        return data

    def get_series_vintage_dates(self, series_id):
        """
        Get a list of vintage dates for a series. Vintage dates are the dates in history when a
        series' data values were revised or new data values were released.

        Parameters
        ----------
        series_id : str
            Fred series id such as 'CPIAUCSL'

        Returns
        -------
        dates : list
            list of vintage dates
        """
        url = "%s/series/vintagedates?series_id=%s" % (self.root_url, series_id)
        root = self.__fetch_data(url)
        if root is None:
            raise ValueError('No vintage date exists for series id: ' + series_id)
        dates = []
        for child in root:
            dates.append(self._parse(child.text))
        return dates

    def __do_series_search(self, url):
        """
        helper function for making one HTTP request for data, and parsing the returned results into a DataFrame
        """
        root = self.__fetch_data(url)

        series_ids = []
        data = {}

        num_results_returned = 0  # number of results returned in this HTTP request
        num_results_total = int(root.get('count'))  # total number of results, this can be larger than number of results returned
        for child in root:
            num_results_returned += 1
            series_id = child.get('id')
            series_ids.append(series_id)
            data[series_id] = {"id": series_id}
            fields = ["realtime_start", "realtime_end", "title", "observation_start", "observation_end",
                      "frequency", "frequency_short", "units", "units_short", "seasonal_adjustment",
                      "seasonal_adjustment_short", "last_updated", "popularity", "notes"]
            for field in fields:
                data[series_id][field] = child.get(field)

        if num_results_returned > 0:
            data = pd.DataFrame(data, columns=series_ids).T
            # parse datetime columns
            for field in ["realtime_start", "realtime_end", "observation_start", "observation_end", "last_updated"]:
                data[field] = data[field].apply(self._parse, format=None)
            # set index name
            data.index.name = 'series id'
        else:
            data = None
        return data, num_results_total

    def __get_search_results(self, url, limit, order_by, sort_order, filter):
        """
        helper function for getting search results up to specified limit on the number of results. The Fred HTTP API
        truncates to 1000 results per request, so this may issue multiple HTTP requests to obtain more available data.
        """

        order_by_options = ['search_rank', 'series_id', 'title', 'units', 'frequency',
                            'seasonal_adjustment', 'realtime_start', 'realtime_end', 'last_updated',
                            'observation_start', 'observation_end', 'popularity']
        if order_by is not None:
            if order_by in order_by_options:
                url = url + '&order_by=' + order_by
            else:
                raise ValueError('%s is not in the valid list of order_by options: %s' % (order_by, str(order_by_options)))

        if filter is not None:
            if len(filter) == 2:
                url = url + '&filter_variable=%s&filter_value=%s' % (filter[0], filter[1])
            else:
                raise ValueError('Filter should be a 2 item tuple like (filter_variable, filter_value)')

        sort_order_options = ['asc', 'desc']
        if sort_order is not None:
            if sort_order in sort_order_options:
                url = url + '&sort_order=' + sort_order
            else:
                raise ValueError('%s is not in the valid list of sort_order options: %s' % (sort_order, str(sort_order_options)))

        data, num_results_total = self.__do_series_search(url)
        if data is None:
            return data

        if limit == 0:
            max_results_needed = num_results_total
        else:
            max_results_needed = limit

        if max_results_needed > self.max_results_per_request:
            for i in range(1, max_results_needed // self.max_results_per_request + 1):
                offset = i * self.max_results_per_request
                next_data, _ = self.__do_series_search(url + '&offset=' + str(offset))
                data = data.append(next_data)
        return data.head(max_results_needed)

    def search(self, text, limit=1000, order_by=None, sort_order=None, filter=None):
        """
        Do a fulltext search for series in the Fred dataset. Returns information about matching series in a DataFrame.

        Parameters
        ----------
        text : str
            text to do fulltext search on, e.g., 'Real GDP'
        limit : int, optional
            limit the number of results to this value. If limit is 0, it means fetching all results without limit.
        order_by : str, optional
            order the results by a criterion. Valid options are 'search_rank', 'series_id', 'title', 'units', 'frequency',
            'seasonal_adjustment', 'realtime_start', 'realtime_end', 'last_updated', 'observation_start', 'observation_end',
            'popularity'
        sort_order : str, optional
            sort the results by ascending or descending order. Valid options are 'asc' or 'desc'
        filter : tuple, optional
            filters the results. Expects a tuple like (filter_variable, filter_value).
            Valid filter_variable values are 'frequency', 'units', and 'seasonal_adjustment'

        Returns
        -------
        info : DataFrame
            a DataFrame containing information about the matching Fred series
        """
        url = "%s/series/search?search_text=%s&" % (self.root_url,
                                                    quote_plus(text))
        info = self.__get_search_results(url, limit, order_by, sort_order, filter)
        return info

    def search_by_release(self, release_id, limit=0, order_by=None, sort_order=None, filter=None):
        """
        Search for series that belongs to a release id. Returns information about matching series in a DataFrame.

        Parameters
        ----------
        release_id : int
            release id, e.g., 151
        limit : int, optional
            limit the number of results to this value. If limit is 0, it means fetching all results without limit.
        order_by : str, optional
            order the results by a criterion. Valid options are 'search_rank', 'series_id', 'title', 'units', 'frequency',
            'seasonal_adjustment', 'realtime_start', 'realtime_end', 'last_updated', 'observation_start', 'observation_end',
            'popularity'
        sort_order : str, optional
            sort the results by ascending or descending order. Valid options are 'asc' or 'desc'
        filter : tuple, optional
            filters the results. Expects a tuple like (filter_variable, filter_value).
            Valid filter_variable values are 'frequency', 'units', and 'seasonal_adjustment'

        Returns
        -------
        info : DataFrame
            a DataFrame containing information about the matching Fred series
        """
        url = "%s/release/series?release_id=%d" % (self.root_url, release_id)
        info = self.__get_search_results(url, limit, order_by, sort_order, filter)
        if info is None:
            raise ValueError('No series exists for release id: ' + str(release_id))
        return info

    def search_by_category(self, category_id, limit=0, order_by=None, sort_order=None, filter=None):
        """
        Search for series that belongs to a category id. Returns information about matching series in a DataFrame.

        Parameters
        ----------
        category_id : int
            category id, e.g., 32145
        limit : int, optional
            limit the number of results to this value. If limit is 0, it means fetching all results without limit.
        order_by : str, optional
            order the results by a criterion. Valid options are 'search_rank', 'series_id', 'title', 'units', 'frequency',
            'seasonal_adjustment', 'realtime_start', 'realtime_end', 'last_updated', 'observation_start', 'observation_end',
            'popularity'
        sort_order : str, optional
            sort the results by ascending or descending order. Valid options are 'asc' or 'desc'
        filter : tuple, optional
            filters the results. Expects a tuple like (filter_variable, filter_value).
            Valid filter_variable values are 'frequency', 'units', and 'seasonal_adjustment'

        Returns
        -------
        info : DataFrame
            a DataFrame containing information about the matching Fred series
        """
        url = "%s/category/series?category_id=%d&" % (self.root_url,
                                                      category_id)
        info = self.__get_search_results(url, limit, order_by, sort_order, filter)
        if info is None:
            raise ValueError('No series exists for category id: ' + str(category_id))
        return info




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_wrds/linking.ipynb.

# %% auto 0
__all__ = ['gvkey_permno_m', 'gvkey_permno_a', 'gvkey_permno_q']

# %% ../../nbs/01_wrds/linking.ipynb 3
import pandas as pd

import pandasmore as pdm
from . import wrds_api

# %% ../../nbs/01_wrds/linking.ipynb 6
def gvkey_permno_m(wrds_username: str=None,
                   obs_limit: int=None, #Number of rows to download. If None, full dataset will be downloaded
                   ) -> pd.DataFrame:
    """CRSP Monthly ids, with gvkeys"""

    limit_clause = f"LIMIT {obs_limit}" if obs_limit is not None else ""
    sql_string=f"""SELECT a.date, a.permno, c.gvkey
                  FROM crsp.msf a
                  INNER JOIN crsp.msenames b ON a.permno = b.permno
                                             AND a.date BETWEEN b.namedt AND b.nameendt 
                  INNER JOIN crsp.ccmxpf_lnkhist c ON a.permno = c.lpermno 
                                                   AND c.linktype IN ('LU','LC') AND c.linkprim IN ('P','C')
                                                   AND a.date BETWEEN c.linkdt AND COALESCE(c.linkenddt, CURRENT_DATE)
                    {limit_clause}                                   
                """
    
    df = wrds_api.download(sql_string, wrds_username)
    df = pdm.setup_panel(df, panel_ids='permno', time_var='date', freq='M',
                         drop_index_duplicates=True, duplicates_which_keep='last')
    df['gvkey'] = df['gvkey'].astype('string')
    return df.reset_index()[['permno','Mdate','gvkey']].copy()  

# %% ../../nbs/01_wrds/linking.ipynb 9
def gvkey_permno_a(wrds_username: str=None,
                   obs_limit: int=None, #Number of rows to download. If None, full dataset will be downloaded
                   ) -> pd.DataFrame:
    """qvkey to permno correspondence at the annual frequency. As done by CCM."""

    limit_clause = f"LIMIT {obs_limit}" if obs_limit is not None else ""
    sql_string=f"""SELECT a.datadate, a.gvkey , b.lpermno as permno
                    FROM comp.funda a
                    INNER JOIN crsp.ccmxpf_lnkhist  b ON a.gvkey = b.gvkey
                    WHERE datadate BETWEEN b.linkdt AND COALESCE(b.linkenddt, CURRENT_DATE)
                            AND b.linktype IN ('LU','LC') AND b.linkprim IN ('P','C')
                            AND indfmt='INDL' AND datafmt='STD' AND popsrc='D' AND consol='C'
                    {limit_clause}            
                """
    
    df = wrds_api.download(sql_string, wrds_username)
    df = pdm.setup_panel(df, panel_ids='permno', time_var='datadate', freq='A',
                         drop_index_duplicates=True, duplicates_which_keep='last')
    df['gvkey'] = df['gvkey'].astype('string')
    return df.reset_index()[['permno','Adate','gvkey']].copy()  

# %% ../../nbs/01_wrds/linking.ipynb 12
def gvkey_permno_q(wrds_username: str=None,
                   obs_limit : int=None, #Number of rows to download. If None, full dataset will be downloaded
                   ) -> pd.DataFrame:
    """qvkey to permno correspondence at the quarterly frequency. As done by CCM."""
    
    limit_clause = f"LIMIT {obs_limit}" if obs_limit is not None else ""
    sql_string=f"""SELECT a.datadate, a.gvkey , b.lpermno as permno
                    FROM comp.fundq a
                    INNER JOIN crsp.ccmxpf_lnkhist  b ON a.gvkey = b.gvkey
                    WHERE datadate BETWEEN b.linkdt AND COALESCE(b.linkenddt, CURRENT_DATE)
                            AND b.linktype IN ('LU','LC') AND b.linkprim IN ('P','C')
                            AND indfmt='INDL' AND datafmt='STD' AND popsrc='D' AND consol='C'            
                    {limit_clause}
                """
    
    df = wrds_api.download(sql_string, wrds_username)
    df = pdm.setup_panel(df, panel_ids='permno', time_var='datadate', freq='Q',
                         drop_index_duplicates=True, duplicates_which_keep='last')
    df['gvkey'] = df['gvkey'].astype('string')
    return df.reset_index()[['permno','Qdate','gvkey']].copy()    



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_wrds/03_compa.ipynb.

# %% ../../nbs/01_wrds/03_compa.ipynb 3
from __future__ import annotations
from pathlib import Path
from typing import List
import os

import pandas as pd
import numpy as np

import pandasmore as pdm
from . import wrds_api
from .. import RESOURCES

# %% auto 0
__all__ = ['raw_metadata', 'raw_metadata_extra', 'default_raw_vars', 'download', 'clean', 'book_equity', 'tobin_q',
           'issuance_vars', 'investment_vars', 'profitability_vars', 'cashflow_vars', 'liquidity_vars', 'leverage_vars',
           'payout_vars']

# %% ../../nbs/01_wrds/03_compa.ipynb 4
def raw_metadata(rawfile: str|Path=RESOURCES/'compa_variable_descriptions.csv', # location of the raw variable labels file
             ) -> pd.DataFrame:
    "Loads raw variable labels file, cleans it and returns it as a pd.DataFrame"

    df = pd.read_csv(rawfile)
    df['output_of'] = 'wrds.compa.clean'

    df['Variable Label'] = df.apply(lambda row: row['Description'].replace(row['Variable Name'].strip()+' -- ', ''), axis=1)
    df['Variable Label'] = df.apply(lambda row: row['Variable Label'].replace( '(' + row['Variable Name'].strip() + ')', ''), axis=1)
    df['Variable Name'] = df['Variable Name'].str.strip().str.lower()
    df = df[['Variable Name', 'Variable Label', 'output_of', 'Type']].copy()
    df.columns = ['name','label','output_of','type']
    return df

# %% ../../nbs/01_wrds/03_compa.ipynb 7
def raw_metadata_extra(wrds_username: str=None
             ) -> pd.DataFrame:
    "Collects metadata from WRDS `comp.funda` table and merges it with `variable_labels`."

    if wrds_username is None:
        wrds_username = os.getenv('WRDS_USERNAME')
        if wrds_username is None: wrds_username = input("Enter your WRDS username: ") 

    try:
        db = wrds_api.Connection(wrds_username = wrds_username)
        funda = db.describe_table('comp','funda')
        nr_rows = db.get_row_count('comp','funda')
    finally:
        db.close()

    meta = funda[['name','type']].copy()
    meta['nr_rows'] = nr_rows
    meta['wrds_library'] = 'comp'
    meta['wrds_table'] = 'funda'

    meta = meta.merge(raw_metadata()[['name','label']], how='left', on='name')
    
    meta['output_of'] = 'wrds.compa.download'
    meta = pdm.order_columns(meta,these_first=['name','label','output_of'])
    for v in list(meta.columns):
        meta[v] = meta[v].astype('string')
    
    return meta

# %% ../../nbs/01_wrds/03_compa.ipynb 9
def default_raw_vars():
    """Default variables used in `download` if none are specified."""

    return ['datadate', 'gvkey', 'cusip' ,'cik' ,'tic' ,'fyear' ,'fyr' ,'naicsh', 'sich' ,'exchg',  
            'lt' ,'at' ,'txditc' ,'pstkl' ,'pstkrv' ,'pstk' ,'csho' ,'ajex' , 'rdip',
            'act' ,'dvc' ,'xad','seq' ,'che' ,'lct' ,'dlc' ,'ib' ,'dvp' ,'txdi' ,'dp' ,
            'txp' ,'oancf' ,'ivncf' ,'fincf' ,'dltt' ,'mib','ceq' ,'invt' ,'cogs' , 'revt',
            'sale' ,'capx' ,'xrd' ,'txdb' ,'prcc_f' ,'sstk' ,'prstkc' ,'dltis' ,'dltr' ,'emp' ,
            'dd1' ,'ppegt' ,'ppent' ,'xint' ,'txt' ,'sppe' ,'gdwl' ,'xrent' ,'re' ,'dvpsx_f' ,
            'tstk' ,'wcap' ,'rect' ,'xsga' ,'aqc' ,'oibdp' ,'dpact' ,'fic' ,'ni' ,'ivao' ,'ivst' ,
            'dv' , 'intan' ,'pi' ,'txfo' ,'pifo' ,'xpp' ,'drc' ,'drlt' ,'ap' ,'xacc' ,'itcb']             

# %% ../../nbs/01_wrds/03_compa.ipynb 11
def download(vars: List[str]=None, # If None, downloads `default_raw_vars`; else `permno`, `permco`, and `date` are added by default
             obs_limit: int=None, #Number of rows to download. If None, full dataset will be downloaded
             wrds_username: str=None, #If None, looks for WRDS_USERNAME with `os.getenv`, then prompts you if needed
             start_date: str="01/01/1900", # Start date in MM/DD/YYYY format
             end_date: str=None #End date in MM/DD/YYYY format; if None, defaults to current date
             ) -> pd.DataFrame:
    """Downloads `vars` from `start_date` to `end_date` from WRDS `comp.funda` library and adds PERMNO and PERMCO as in CCM"""

    if vars is None: vars = default_raw_vars()
    vars = ','.join(['a.gvkey', 'a.datadate'] + 
                    [f'a.{x}' for x in vars if x not in ['datadate', 'gvkey']])

    limit_clause = f"LIMIT {obs_limit}" if obs_limit is not None else ""
    sql_string=f"""SELECT b.lpermno as permno, b.lpermco as permco, b.liid as iid, {vars}
                    FROM comp.funda AS a
                    INNER JOIN crsp.ccmxpf_lnkhist AS b ON a.gvkey = b.gvkey
                    WHERE datadate BETWEEN b.linkdt AND COALESCE(b.linkenddt, CURRENT_DATE)
                            AND b.linktype IN ('LU','LC') AND b.linkprim IN ('P','C')
                            AND indfmt='INDL' AND datafmt='STD' AND popsrc='D' AND consol='C'
                            AND datadate BETWEEN '{start_date}' AND COALESCE(%(end)s, CURRENT_DATE)
                    {limit_clause}
                """
    return wrds_api.download(sql_string, wrds_username=wrds_username, params={'end':end_date})

# %% ../../nbs/01_wrds/03_compa.ipynb 14
def clean(df: pd.DataFrame=None,        # If None, downloads `vars` using `download` function; else, must contain `permno` and `datadate` columns
          vars: List[str]=None,         # If None, downloads `default_raw_vars`
          obs_limit: int=None, #Number of rows to download. If None, full dataset will be downloaded
          wrds_username: str=None,      # If None, looks for WRDS_USERNAME with `os.getenv`, then prompts you if needed
          start_date: str="01/01/1900", # Start date in MM/DD/YYYY format
          end_date: str=None,           # End date. Default is current date          
          clean_kwargs: dict={},        # Params to pass to `pdm.setup_panel` other than `panel_ids`, `time_var`, and `freq`
          ) -> pd.DataFrame:
    """Applies `pandasmore.setup_panel` to `df`. If `df` is None, downloads `vars` using `download` function."""

    if df is None: df = download(vars=vars, obs_limit=obs_limit,  wrds_username=wrds_username, start_date=start_date, end_date=end_date)
    df = pdm.setup_panel(df, panel_ids='permno', time_var='datadate', freq='Y', **clean_kwargs)
    return df 

# %% ../../nbs/01_wrds/03_compa.ipynb 19
def book_equity(df: pd.DataFrame=None, # If None, downloads (and cleans) only required vars
                add_itcb=False,
                return_metadata: bool=False # If true, just returns metadata dictionary
                ) -> pd.DataFrame:

    metadata = {'inputs': {'wrds.compa.clean': ['at', 'lt', 'seq', 'ceq', 'txditc', 'pstk', 'pstkrv', 'pstkl', 'itcb']},
                'outputs': ['bookeq','shreq','pref_stock'],
                'labels': {'bookeq': 'Book equity', 'shreq': 'Shareholder equity', 'pref_stock': 'Preferred stock'}
    }      
    if return_metadata: return metadata

    reqs = metadata['inputs']['wrds.compa.clean']
    if df is None: df = clean(vars=reqs)
    df = df[reqs].copy()

    df['pstk'] = df['pstk'].fillna(0)
    df['pref_stock'] = np.where(df['pstkrv'].isnull(), df['pstkl'], df['pstkrv'])
    df['pref_stock'] = np.where(df['pref_stock'].isnull(),df['pstk'], df['pref_stock'])

    df['shreq'] = np.where(df['seq'].isnull(), df['ceq'] + df['pstk'], df['seq'])
    df['shreq'] = np.where(df['shreq'].isnull(), df['at'] - df['lt'], df['shreq'])

    df['bookeq'] = df['shreq'] + df['txditc'].fillna(0) - df['pref_stock']
    if add_itcb: df['bookeq'] = df['bookeq'] + df['itcb'].fillna(0)
    
    return df[metadata['outputs']].copy()

# %% ../../nbs/01_wrds/03_compa.ipynb 26
def tobin_q(df: pd.DataFrame=None,      # If None, downloads (and cleans) only required vars
            return_metadata: bool=False # If True, just returns the metadata dictionary
            ) -> pd.DataFrame:

    metadata = {'inputs': {'wrds.compa.clean': ['at', 'lt', 'seq', 'ceq', 'txditc', 'pstk', 'pstkrv', 'pstkl', 'itcb','prcc_f','csho']},
                'outputs':  ['tobinq'],
                'labels': {'tobinq': 'Tobin Q'}
    }      
    if return_metadata: return metadata

    reqs = metadata['inputs']['wrds.compa.clean']
    if df is None: df = clean(vars=reqs)
    df = df[reqs].copy()

    beq = book_equity(df)[['bookeq']].copy()
    df = df.join(beq)

    df['tobinq'] = (df['at'] - df['bookeq'] + df['prcc_f'] * df['csho']) / df['at']
    df = df.replace([np.inf, -np.inf], np.nan)
    return  df[metadata['outputs']].copy()

# %% ../../nbs/01_wrds/03_compa.ipynb 31
def issuance_vars(df: pd.DataFrame=None,        # If None, downloads (and cleans) only required vars
                    return_metadata: bool=False # If True, just returns the metadata dictionary
                    ) -> pd.DataFrame:

    metadata = {'inputs': {'wrds.compa.clean': ['at', 'lt', 'seq', 'ceq', 'txditc', 'pstk', 'pstkrv', 'pstkl', 'itcb',
                                                  'sstk','prstkc','dltis','dltr', 're', 'dlc','dltt']},
                'outputs': ['equityiss_tot','equityiss_cfs', 'debtiss_tot', 'debtiss_cfs', 'debtiss_bs'],
                'labels': {'equityiss_tot':'Equity issuance','equityiss_cfs':'Equity issuance', 
                           'debtiss_tot':'Debt issuance', 'debtiss_cfs':'Debt issuance', 'debtiss_bs':'Debt issuance'}
    }      
    if return_metadata: return metadata

    reqs = metadata['inputs']['wrds.compa.clean']
    if df is None: df = clean(vars=reqs)
    df = df[reqs].copy()

    beq = book_equity(df)[['bookeq']].copy()
    df = df.join(beq)
    
    df['lag_at'] = pdm.lag(df['at'])

    df['equityiss_cfs'] = (df['sstk'].fillna(0) - df['prstkc'].fillna(0)) / df['lag_at']
    df['debtiss_cfs'] = (df['dltis'].fillna(0) - df['dltr'].fillna(0)) / df['lag_at']

    df['debtiss_bs'] = (pdm.rdiff(df['dltt']) + pdm.rdiff(df['dlc'].fillna(0))) / df['lag_at']

    df['equityiss_tot'] = (pdm.rdiff(df['bookeq']) - pdm.rdiff(df['re'])) / df['lag_at']
    df['debtiss_tot'] = (pdm.rdiff(df['at']) - pdm.rdiff(df['bookeq'])) / df['lag_at']
    
    df = df.replace([np.inf, -np.inf], np.nan)
    return df[metadata['outputs']].copy()

# %% ../../nbs/01_wrds/03_compa.ipynb 34
def investment_vars(df: pd.DataFrame=None,      # If None, downloads (and cleans) only required vars
                    return_metadata: bool=False # If True, just returns the metadata dictionary
                    ) -> pd.DataFrame:

    metadata = {'inputs': {'wrds.compa.clean': ['ppent','capx','at']},
                'outputs': ['ppentpch','capx2la'],
                'labels': {'ppentpch':'Pct change in net PPE','capx2la': 'CAPX to lagged assets'}
    }      
    if return_metadata: return metadata

    reqs = metadata['inputs']['wrds.compa.clean']
    if df is None: df = clean(vars=reqs)
    df = df[reqs].copy()
    
    df['ppentpch'] = pdm.rpct_change(df['ppent'])
    df['capx2la'] = df['capx'] / pdm.lag(df['at'])
    df = df.replace([np.inf, -np.inf], np.nan)
    return df[metadata['outputs']].copy()

# %% ../../nbs/01_wrds/03_compa.ipynb 37
def profitability_vars(df: pd.DataFrame=None,      # If None, downloads (and cleans) only required vars
                        return_metadata: bool=False # If True, just returns the metadata dictionary
                        ) -> pd.DataFrame:

    metadata = {'inputs': {'wrds.compa.clean': ['ib','at']},
                'outputs': ['roa'],
                'labels': {'roa':'Return on assets'}
    }      
    if return_metadata: return metadata

    reqs = metadata['inputs']['wrds.compa.clean']
    if df is None: df = clean(vars=reqs)
    df = df[reqs].copy()

    df['roa'] = df['ib'] / df['at']
    df = df.replace([np.inf, -np.inf], np.nan)
    return df[metadata['outputs']].copy()

# %% ../../nbs/01_wrds/03_compa.ipynb 40
def cashflow_vars(df: pd.DataFrame=None,      # If None, downloads (and cleans) only required vars
                        return_metadata: bool=False # If True, just returns the metadata dictionary
                        ) -> pd.DataFrame:

    metadata = {'inputs': {'wrds.compa.clean': ['dtdate','oancf','ib','dp','at']},
                'outputs': ['cflow2la_is', 'cflow2la_cfs', 'cflow2la_full'],
                'labels': {'cflow2la_is':'Cash flows to lagged assets', 
                           'cflow2la_cfs':'Cash flows to lagged assets', 
                           'cflow2la_full':'Cash flows to lagged assets'}
    }      
    if return_metadata: return metadata

    reqs = metadata['inputs']['wrds.compa.clean']
    if df is None: df = clean(vars=reqs)
    df = df[reqs].copy()
    
    df['cflow2la_is'] = (df['ib']+df['dp']) / pdm.lag(df['at'])
    df['cflow2la_cfs'] = df['oancf'] / pdm.lag(df['at'])
    df['cflow2la_full'] = np.where(df.dtdate.dt.year<1987, df['cflow2la_is'], df['cflow2la_cfs'])

    df = df.replace([np.inf, -np.inf], np.nan)    
    return df[metadata['outputs']].copy()

# %% ../../nbs/01_wrds/03_compa.ipynb 43
def liquidity_vars(df: pd.DataFrame=None,      # If None, downloads (and cleans) only required vars
                        return_metadata: bool=False # If True, just returns the metadata dictionary
                        ) -> pd.DataFrame:

    metadata = {'inputs': {'wrds.compa.clean': ['che','at']},
                'outputs': ['cash2a'],
                'labels': {'cash2a':'Cash holdings to assets'}
    }      
    if return_metadata: return metadata

    reqs = metadata['inputs']['wrds.compa.clean']
    if df is None: df = clean(vars=reqs)
    df = df[reqs].copy()

    df['cash2a'] = df['che'] / df['at']

    df = df.replace([np.inf, -np.inf], np.nan) 
    return df[metadata['outputs']].copy()

# %% ../../nbs/01_wrds/03_compa.ipynb 46
def leverage_vars(df: pd.DataFrame=None,      # If None, downloads (and cleans) only required vars
                        return_metadata: bool=False # If True, just returns the metadata dictionary
                        ) -> pd.DataFrame:

    metadata = {'inputs': {'wrds.compa.clean': ['dltt','dlc','at']},
                'outputs': ['booklev'],
                'labels': {'booklev':'Book leverage'}
    }      
    if return_metadata: return metadata

    reqs = metadata['inputs']['wrds.compa.clean']
    if df is None: df = clean(vars=reqs)
    df = df[reqs].copy()

    df['booklev'] = (df['dltt'] + df['dlc']) / df['at']
    df.loc[df.booklev<0, 'booklev'] = 0
    df.loc[df.booklev>1, 'booklev'] = 1
    
    df = df.replace([np.inf, -np.inf], np.nan)         
    return df[metadata['outputs']].copy()

# %% ../../nbs/01_wrds/03_compa.ipynb 49
def payout_vars(df: pd.DataFrame=None,      # If None, downloads (and cleans) only required vars
                        return_metadata: bool=False # If True, just returns the metadata dictionary
                        ) -> pd.DataFrame:

    metadata = {'inputs': {'wrds.compa.clean': ['dvc','prstkc','at']},
                'outputs': ['div2la','rep2la'],
                'labels': {'div2la': 'Dividends to lagged assets',
                           'rep2la': 'Repurchases to lagged assets'}
    }      
    if return_metadata: return metadata

    reqs = metadata['inputs']['wrds.compa.clean']
    if df is None: df = clean(vars=reqs)
    df = df[reqs].copy()

    df['div2la'] = df['dvc'].fillna(0) / pdm.lag(df['at'])
    df['rep2la'] = df['prstkc'].fillna(0) / pdm.lag(df['at'])

    df = df.replace([np.inf, -np.inf], np.nan) 
    return df[metadata['outputs']].copy()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_wrds/00_wrds_api.ipynb.

# %% auto 0
__all__ = ['wrds_version', 'appname', 'WRDS_POSTGRES_HOST', 'WRDS_POSTGRES_PORT', 'WRDS_POSTGRES_DB', 'WRDS_CONNECT_ARGS',
           'NotSubscribedError', 'SchemaNotFoundError', 'Connection', 'download']

# %% ../../nbs/01_wrds/00_wrds_api.ipynb 3
from typing import Sequence
import getpass
import os, glob
import sys
import stat
import pandas as pd
import sqlalchemy as sa
import urllib.parse
from sys import version_info

# %% ../../nbs/01_wrds/00_wrds_api.ipynb 4
wrds_version = '3.1.16'

appname = "{0} python {1}.{2}.{3}/wrds {4}".format(
    sys.platform, version_info[0], version_info[1], version_info[2], wrds_version
)

# Sane defaults
WRDS_POSTGRES_HOST = "wrds-pgdata.wharton.upenn.edu"
WRDS_POSTGRES_PORT = 9737
WRDS_POSTGRES_DB = "wrds"
WRDS_CONNECT_ARGS = {"sslmode": "require", "application_name": appname}

class NotSubscribedError(PermissionError):
    pass

class SchemaNotFoundError(FileNotFoundError):
    pass

# %% ../../nbs/01_wrds/00_wrds_api.ipynb 5
class Connection(object):
    def __init__(self, autoconnect=True, verbose=False, **kwargs):
        """
        Set up the connection to the WRDS database.
        By default, also establish the connection to the database.

        Optionally, the user may specify connection parameters:
            *wrds_hostname*: WRDS database hostname
            *wrds_port*: database connection port number
            *wrds_dbname*: WRDS database name
            *wrds_username*: WRDS username
            *autoconnect*: If false will not immediately establish the connection

        The constructor will use the .pgpass file if it exists and may make use of
        PostgreSQL environment variables such as PGHOST, PGUSER, etc., if cooresponding
        parameters are not set.
        If not, it will ask the user for a username and password.
        It will also direct the user to information on setting up .pgpass.

        Additionally, creating the instance will load a list of schemas
        the user has permission to access.

        :return: None

        Usage::
        >>> db = wrds.Connection()
        Loading library list...
        Done
        """
        self._verbose = verbose
        self._password = os.getenv("WRDS_PASS","") # MION: originally this was self._password = ""
        # If user passed in any of these parameters, override defaults.
        self._username = kwargs.get("wrds_username", "")
        # PGHOST if set will override default for first attempt
        self._hostname = kwargs.get(
            "wrds_hostname", os.environ.get('PGHOST', WRDS_POSTGRES_HOST)
        )
        self._port = kwargs.get("wrds_port", WRDS_POSTGRES_PORT)
        self._dbname = kwargs.get("wrds_dbname", WRDS_POSTGRES_DB)
        self._connect_args = kwargs.get("wrds_connect_args", WRDS_CONNECT_ARGS)

        if autoconnect:
            self.connect()
            self.load_library_list()

    def __make_sa_engine_conn(self, raise_err=True):  #MION changed raise_err default to True
        username = self._username
        hostname = self._hostname
        password = urllib.parse.quote_plus(self._password)
        port = self._port
        dbname = self._dbname
        pguri = f"postgresql://{username}:{password}@{hostname}:{port}/{dbname}"
        if self._verbose:
            print(f"postgresql://{username}:@{hostname}:{port}/{dbname}")
        try:
            self.engine = sa.create_engine(
                pguri,
                isolation_level="AUTOCOMMIT",
                connect_args=self._connect_args,
            )
            self.connection = self.engine.connect()
        except Exception as err:
            if self._verbose:
                print(f"{err=}")
            self.engine = None
            if raise_err:
                raise err

    def connect(self):
        """Make a connection to the WRDS database."""
        # first try connection using system defaults and params set in constructor
        self.__make_sa_engine_conn()

        if (self.engine is None and self._hostname != WRDS_POSTGRES_HOST):
            # try explicit w/ default hostname
            print(f"Trying '{WRDS_POSTGRES_HOST}'...")
            self._hostname = WRDS_POSTGRES_HOST
            self.__make_sa_engine_conn()

        if (self.engine is None):
            # Use explicit username and password
            self._username, self._password = self.__get_user_credentials()
            # Last attempt, raise error if Exception encountered
            self.__make_sa_engine_conn(raise_err=True)

            if (self.engine is None):
                print(f"Failed to connect {self._username}@{self._hostname}")
            else:
                # Connection successful. Offer to create a .pgpass for the user.
                print("WRDS recommends setting up a .pgpass file.")
                do_create_pgpass = ""
                while do_create_pgpass != "y" and do_create_pgpass != "n":
                    do_create_pgpass = input("Create .pgpass file now [y/n]?: ")

                if do_create_pgpass == "y":
                    try:
                        self.create_pgpass_file()
                        print("Created .pgpass file successfully.")
                    except Exception:
                        print("Failed to create .pgpass file.")
                print(
                    "You can create this file yourself at any time "
                    "with the create_pgpass_file() function."
                )

    def close(self):
        """
        Close the connection to the database.
        """
        self.connection.close()
        self.engine.dispose()
        self.engine = None

    def __enter__(self):
        print("Making connection")
        self.connect()
        return self

    def __exit__(self, *args):
        print("closing connection")
        self.close()

    def load_library_list(self):
        """Load the list of Postgres schemata (c.f. SAS LIBNAMEs)
        the user has permission to access."""
        self.insp = sa.inspect(self.connection)
        print("Loading library list...")
        query = """
WITH pgobjs AS (
    -- objects we care about - tables, views, foreign tables, partitioned tables
    SELECT oid, relnamespace, relkind
    FROM pg_class
    WHERE relkind = ANY (ARRAY['r'::"char", 'v'::"char", 'f'::"char", 'p'::"char"])
),
schemas AS (
    -- schemas we have usage on that represent products
    SELECT nspname AS schemaname,
        pg_namespace.oid,
        array_agg(DISTINCT relkind) AS relkind_a
    FROM pg_namespace
    JOIN pgobjs ON pg_namespace.oid = relnamespace
    WHERE nspname !~ '(^pg_)|(_old$)|(_new$)|(information_schema)'
        AND has_schema_privilege(nspname, 'USAGE') = TRUE
    GROUP BY nspname, pg_namespace.oid
)
SELECT schemaname
FROM schemas
WHERE relkind_a != ARRAY['v'::"char"] -- any schema except only views
UNION
-- schemas w/ views (aka "friendly names") that reference accessable product tables
SELECT nv.schemaname
FROM schemas nv
JOIN pgobjs v ON nv.oid = v.relnamespace AND v.relkind = 'v'::"char"
JOIN pg_depend dv ON v.oid = dv.refobjid AND dv.refclassid = 'pg_class'::regclass::oid
    AND dv.classid = 'pg_rewrite'::regclass::oid AND dv.deptype = 'i'::"char"
JOIN pg_depend dt ON dv.objid = dt.objid AND dv.refobjid <> dt.refobjid
    AND dt.classid = 'pg_rewrite'::regclass::oid
    AND dt.refclassid = 'pg_class'::regclass::oid
JOIN pgobjs t ON dt.refobjid = t.oid
    AND (t.relkind = ANY (ARRAY['r'::"char", 'v'::"char", 'f'::"char", 'p'::"char"]))
JOIN schemas nt ON t.relnamespace = nt.oid
GROUP BY nv.schemaname
ORDER BY 1;
        """
        cursor = self.connection.execute(query)
        self.schema_perm = [x[0] for x in cursor.fetchall()]
        print("Done")

    def __get_user_credentials(self):
        """Prompt the user for their WRDS credentials.

        Use the OS-level username as a default so the user
          doesn't have to reenter it if they match.
        Return both the username and the password.

        >>> user,passwd = wrds.Connection.__get_user_credentials()
        """
        if self._username:
            uname = self._username
        else:
            uname = getpass.getuser()
        username = input("Enter your WRDS username [{}]:".format(uname))
        if not username:
            username = uname
        passwd = getpass.getpass("Enter your password:")
        return username, passwd

    def create_pgpass_file(self):
        """
        Create a .pgpass file to store WRDS connection credentials.

        Use the existing username and password if already connected to WRDS,
         or prompt for that information if not.

        The .pgpass file may contain connection entries for multiple databases,
          so we take care not to overwrite any existing entries unless they
          have the same hostname, port, and database name.

        On Windows, this file is actually called "pgpass.conf"
          and is stored in the %APPDATA%\postgresql directory.
        This must be handled differently.

        Usage:
        >>> db = wrds.Connection()
        >>> db.create_pgpass_file()
        """
        if not self._username or not self._password:
            self._username, self._password = self.__get_user_credentials()
        if sys.platform == "win32":
            self.__create_pgpass_file_win32()
        else:
            self.__create_pgpass_file_unix()

    def __create_pgpass_file_win32(self):
        """
        Create a pgpass.conf file on Windows.

        Windows is different enough from everything else
          as to require its own special way of doing things.
        Save the pgpass file in %APPDATA%\postgresql as 'pgpass.conf'.
        """
        appdata = os.getenv("APPDATA")
        pgdir = appdata + os.path.sep + "postgresql"
        # Can we at least assume %APPDATA% always exists? I'm seriously asking.
        if not os.path.exists(pgdir):
            os.mkdir(pgdir)
        # Path exists, but is not a directory
        elif not os.path.isdir(pgdir):
            err = "Cannot create directory {}: " "path exists but is not a directory"
            raise FileExistsError(err.format(pgdir))
        pgfile = pgdir + os.path.sep + "pgpass.conf"
        # Write the pgpass.conf file without clobbering
        self.__write_pgpass_file(pgfile)

    def __create_pgpass_file_unix(self):
        """
        Create a .pgpass file on Unix-like operating systems.

        Permissions on this file must also be set on Unix-like systems.
        This function works on Mac OS X and Linux.
        It should work on Solaris too, but this is untested.
        """
        homedir = os.getenv("HOME")
        pgfile = homedir + os.path.sep + ".pgpass"
        if os.path.isfile(pgfile):
            # Set it to mode 600 (rw-------) so we can write to it
            os.chmod(pgfile, stat.S_IRUSR | stat.S_IWUSR)
        self.__write_pgpass_file(pgfile)
        # Set it to mode 400 (r------) to protect it
        os.chmod(pgfile, stat.S_IRUSR)

    def __write_pgpass_file(self, pgfile):
        """
        Write the WRDS connection info to the pgpass file
          without clobbering other connection strings.

        Also escape any ':' characters in passwords,
          as .pgpass requires.

        Works on both *nix and Win32.
        """
        pgpass = "{host}:{port}:{dbname}:{user}:{passwd}\n"
        passwd = self._password
        passwd = passwd.replace(":", "\:")
        # Avoid clobbering the file if it exists
        if os.path.isfile(pgfile):
            with open(pgfile, "r") as fd:
                lines = fd.readlines()
            newlines = []
            for line in lines:
                # Handle escaped colons, preventing
                #  split() from splitting on them.
                # Saving to a new variable here absolves us
                #  of having to re-replace the substituted ##COLON## later.
                oldline = line.replace("""\:""", "##COLON##")
                fields = oldline.split(":")
                # On finding a line matching the hostname, port and dbname
                #  we replace it with the new pgpass line.
                # Surely we won't have any colons in these fields :^)
                if (
                    fields[0] == self._hostname
                    and int(fields[1]) == self._port
                    and fields[2] == self._dbname
                    and fields[3] == self._username
                ):
                    newline = pgpass.format(
                        host=self._hostname,
                        port=self._port,
                        dbname=self._dbname,
                        user=self._username,
                        passwd=passwd,
                    )
                    newlines.append(newline)
                else:
                    newlines.append(line)

            # Add line for current user/password - enables multiple wrds-pgdata entries
            # with different usernames
            newline = pgpass.format(
                host=self._hostname,
                port=self._port,
                dbname=self._dbname,
                user=self._username,
                passwd=passwd,
            )
            if newline not in newlines:
                newlines.append(newline)
            lines = newlines
        else:
            line = pgpass.format(
                host=self._hostname,
                port=self._port,
                dbname=self._dbname,
                user=self._username,
                passwd=passwd,
            )
            lines = [line]
        # I lied, we're totally clobbering it:
        with open(pgfile, "w") as fd:
            fd.writelines(lines)

    def __check_schema_perms(self, schema):
        """
        Check the permissions of the schema.
        Raise permissions error if user does not have access.
        Raise other error if the schema does not exist.

        Else, return True

        :param schema: Postgres schema name.
        :rtype: bool

        """
        if schema in self.schema_perm:
            return True
        else:
            if schema in self.insp.get_schema_names():
                raise NotSubscribedError(
                    "You do not have permission to access "
                    "the {} library".format(schema)
                )
            else:
                raise SchemaNotFoundError("The {} library is not found.".format(schema))

    def list_libraries(self):
        """
        Return all the libraries (schemas) the user can access.

        :rtype: list

        Usage::
        >>> db.list_libraries()
        ['aha', 'audit', 'block', 'boardex', ...]
        """
        return self.schema_perm

    def list_tables(self, library):
        """
        Returns a list of all the views/tables/foreign tables within a schema.

        :param library: Postgres schema name.

        :rtype: list

        Usage::
        >>> db.list_tables('wrdssec')
        ['wciklink_gvkey', 'dforms', 'wciklink_cusip', 'wrds_forms', ...]
        """
        if self.__check_schema_perms(library):
            output = (
                self.insp.get_view_names(schema=library)
                + self.insp.get_table_names(schema=library)
                + self.insp.get_foreign_table_names(schema=library)
            )
            return output

    def __get_schema_for_view(self, schema, table):
        """
        Internal function for getting the schema based on a view
        """
        sql_code = """SELECT distinct(source_ns.nspname) AS source_schema
                      FROM pg_depend
                      JOIN pg_rewrite
                        ON pg_depend.objid = pg_rewrite.oid
                      JOIN pg_class as dependent_view
                        ON pg_rewrite.ev_class = dependent_view.oid
                      JOIN pg_class as source_table
                        ON pg_depend.refobjid = source_table.oid
                      JOIN pg_attribute
                        ON pg_depend.refobjid = pg_attribute.attrelid
                          AND pg_depend.refobjsubid = pg_attribute.attnum
                      JOIN pg_namespace dependent_ns
                        ON dependent_ns.oid = dependent_view.relnamespace
                      JOIN pg_namespace source_ns
                        ON source_ns.oid = source_table.relnamespace
                      WHERE dependent_ns.nspname = '{schema}'
                        AND dependent_view.relname = '{view}';
                    """.format(
            schema=schema, view=table
        )
        if self.__check_schema_perms(schema):
            result = self.connection.execute(sql_code)
            return result.fetchone()[0]

    def describe_table(self, library, table):
        """
        Takes the library and the table and describes all the columns
          in that table.
        Includes Column Name, Column Type, Nullable?, Comment

        :param library: Postgres schema name.
        :param table: Postgres table name.

        :rtype: pandas.DataFrame

        Usage::
        >>> db.describe_table('wrdssec_all', 'dforms')
                    name nullable     type comment
              0      cik     True  VARCHAR
              1    fdate     True     DATE
              2  secdate     True     DATE
              3     form     True  VARCHAR
              4   coname     True  VARCHAR
              5    fname     True  VARCHAR
        """
        rows = self.get_row_count(library, table)
        print("Approximately {} rows in {}.{}.".format(rows, library, table))
        table_info = pd.DataFrame.from_dict(
            self.insp.get_columns(table, schema=library)
        )
        return table_info[["name", "nullable", "type", "comment"]]

    def get_row_count(self, library, table):
        """
        Uses the library and table to get the approximate row count for the table.

        :param library: Postgres schema name.
        :param table: Postgres table name.

        :rtype: int

        Usage::
        >>> db.get_row_count('wrdssec', 'dforms')
        16378400
        """

        sqlstmt = """
            EXPLAIN (FORMAT 'json')  SELECT 1 FROM {}.{} ;
        """.format(
            sa.sql.quoted_name(library, True), sa.sql.quoted_name(table, True)
        )

        try:
            result = self.connection.execute(sqlstmt)
            return int(result.fetchone()[0][0]["Plan"]["Plan Rows"])
        except Exception as e:
            print("There was a problem with retrieving the row count: {}".format(e))
            return 0

    def raw_sql(
        self,
        sql,
        coerce_float=True,
        date_cols=None,
        index_col=None,
        params=None,
        chunksize=500000,
        return_iter=False,
    ):
        """
        Queries the database using a raw SQL string.

        :param sql: SQL code in string object.
        :param coerce_float: (optional) boolean, default: True
            Attempt to convert values to non-string, non-numeric objects
            to floating point. Can result in loss of precision.
        :param date_cols: (optional) list or dict, default: None
            - List of column names to parse as date
            - Dict of ``{column_name: format string}`` where
                format string is:
                  strftime compatible in case of parsing string times or
                  is one of (D, s, ns, ms, us) in case of parsing
                    integer timestamps
            - Dict of ``{column_name: arg dict}``,
                where the arg dict corresponds to the keyword arguments of
                  :func:`pandas.to_datetime`
        :param index_col: (optional) string or list of strings,
          default: None
            Column(s) to set as index(MultiIndex)
        :param params: parameters to SQL query, if parameterized.
        :param chunksize: (optional) integer or None default: 500000
            Process query in chunks of this size. Smaller chunksizes can save
            a considerable amount of memory while query is being processed.
            Set to None run query w/o chunking.
        :param return_iter: (optional) boolean, default:False
            When chunksize is not None, return an iterator where chunksize
            number of rows is included in each chunk.

        :rtype: pandas.DataFrame or or Iterator[pandas.DataFrame]


        Usage ::
        # Basic Usage
        >>> data = db.raw_sql('select cik, fdate, coname from wrdssec_all.dforms;', date_cols=['fdate'], index_col='cik')
        >>> data.head()
            cik        fdate       coname
            0000000003 1995-02-15  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
            0000000003 1996-02-14  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
            0000000003 1997-02-19  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
            0000000003 1998-03-02  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
            0000000003 1998-03-10  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y..
            ...

        # Parameterized SQL query
        >>> parm = {'syms': ('A', 'AA', 'AAPL'), 'num_shares': 50000}
        >>> data = db.raw_sql('select * from taqmsec.ctm_20030910 where sym_root in %(syms)s and size > %(num_shares)s', params=parm)
        >>> data.head()
                  date           time_m ex sym_root sym_suffix tr_scond      size   price tr_stopind tr_corr     tr_seqnum tr_source tr_rf
            2003-09-10  11:02:09.485000  T        A       None     None  211400.0  25.350          N      00  1.929952e+15         C  None
            2003-09-10  11:04:29.508000  N        A       None     None   55500.0  25.180          N      00  1.929952e+15         C  None
            2003-09-10  15:08:21.155000  N        A       None     None   50500.0  24.470          N      00  1.929967e+15         C  None
            2003-09-10  16:10:35.522000  T        A       None        B   71900.0  24.918          N      00  1.929970e+15         C  None
            2003-09-10  09:35:20.709000  N       AA       None     None  108100.0  28.200          N      00  1.929947e+15         C  None
        """  # noqa

        try:
            df = pd.read_sql_query(
                sql,
                self.connection,
                coerce_float=coerce_float,
                parse_dates=date_cols,
                index_col=index_col,
                chunksize=chunksize,
                params=params,
            )
            if return_iter or chunksize is None:
                return df
            else:
                full_df = pd.DataFrame()
                for chunk in df:
                    full_df = pd.concat([full_df, chunk])
                return full_df
        except sa.exc.ProgrammingError as e:
            raise e

    def get_table(
        self,
        library,
        table,
        obs=-1,
        offset=0,
        columns=None,
        coerce_float=None,
        index_col=None,
        date_cols=None,
    ):
        """
        Creates a data frame from an entire table in the database.

        :param sql: SQL code in string object.
        :param library: Postgres schema name.

        :param obs: (optional) int, default: -1
            Specifies the number of observations to pull from the table.
            An integer less than 0 will return the entire table.
        :param offset: (optional) int, default: 0
            Specifies the starting point for the query.
            An offset of 0 will start selecting from the beginning.
        :param columns: (optional) list or tuple, default: None
            Specifies the columns to be included in the output data frame.
        :param coerce_float: (optional) boolean, default: True
            Attempt to convert values to non-string, non-numeric objects
            to floating point. Can result in loss of precision.
        :param date_cols: (optional) list or dict, default: None
            - List of column names to parse as date
            - Dict of ``{column_name: format string}``
                where format string is
                  strftime compatible in case of parsing string times or
                  is one of (D, s, ns, ms, us) in case of parsing
                    integer timestamps
            - Dict of ``{column_name: arg dict}``,
                where the arg dict corresponds to the keyword arguments of
                  :func:`pandas.to_datetime`
        :param index_col: (optional) string or list of strings,
          default: None
            Column(s) to set as index(MultiIndex)

        :rtype: pandas.DataFrame

        Usage ::
        >>> data = db.get_table('wrdssec_all', 'dforms', obs=1000, columns=['cik', 'fdate', 'coname'])
        >>> data.head()
            cik        fdate       coname
            0000000003 1995-02-15  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
            0000000003 1996-02-14  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
            0000000003 1997-02-19  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
            0000000003 1998-03-02  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y...
            0000000003 1998-03-10  DEFINED ASSET FUNDS MUNICIPAL INVT TR FD NEW Y..
            ...

        """  # noqa
        if obs < 0:
            obsstmt = ""
        else:
            obsstmt = " LIMIT {}".format(obs)
        if columns is None:
            cols = "*"
        else:
            cols = ",".join(columns)
        if self.__check_schema_perms(library):
            sqlstmt = (
                "SELECT {cols} FROM {schema}.{table} {obsstmt} OFFSET {offset};".format(
                    cols=cols,
                    schema=library,
                    table=table,
                    obsstmt=obsstmt,
                    offset=offset,
                )
            )
            return self.raw_sql(
                sqlstmt,
                coerce_float=coerce_float,
                index_col=index_col,
                date_cols=date_cols,
            )


# %% ../../nbs/01_wrds/00_wrds_api.ipynb 6
def download(sql_string: str=None,
             wrds_username: str=None, #If None, looks for WRDS_USERNAME with `os.getenv`; prompts you if it can't find it
             params: Sequence=None # Params cited in the `sql_string`
             ) -> pd.DataFrame:
    """Downloads data from WRDS using the given PostgreSQL `sql_string`"""

    if wrds_username is None:
        wrds_username = os.getenv('WRDS_USERNAME')
        if wrds_username is None: wrds_username = input("Enter your WRDS username: ") 

    #The context manager does not seem to close connections properly
    #with Connection(wrds_username = wrds_username) as db:
    #    df = db.raw_sql(sql=sql_string, params=params)

    try:
        db = Connection(wrds_username=os.getenv('WRDS_USERNAME'))
        df = db.raw_sql(sql=sql_string, params=params)
    except Exception as err:
        raise err 
    finally:
        db.close()

    return df



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/finsets',
                'doc_host': 'https://ionmihai.github.io',
                'git_url': 'https://github.com/ionmihai/finsets',
                'lib_path': 'finsets'},
  'syms': { 'finsets.cli': {'finsets.cli.search': ('cli.html#search', 'finsets/cli.py')},
            'finsets.core': { 'finsets.core.all_metadata': ('core.html#all_metadata', 'finsets/core.py'),
                              'finsets.core.features_metadata': ('core.html#features_metadata', 'finsets/core.py'),
                              'finsets.core.raw_metadata': ('core.html#raw_metadata', 'finsets/core.py'),
                              'finsets.core.search': ('core.html#search', 'finsets/core.py')},
            'finsets.fred.fred': { 'finsets.fred.fred.clean': ('00_fred/fred.html#clean', 'finsets/fred/fred.py'),
                                   'finsets.fred.fred.common_raw_vars': ('00_fred/fred.html#common_raw_vars', 'finsets/fred/fred.py'),
                                   'finsets.fred.fred.download': ('00_fred/fred.html#download', 'finsets/fred/fred.py'),
                                   'finsets.fred.fred.get_series_info': ('00_fred/fred.html#get_series_info', 'finsets/fred/fred.py'),
                                   'finsets.fred.fred.search': ('00_fred/fred.html#search', 'finsets/fred/fred.py')},
            'finsets.fred.fred_api': { 'finsets.fred.fred_api.Fred': ('00_fred/fred_api.html#fred', 'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.__do_series_search': ( '00_fred/fred_api.html#fred.__do_series_search',
                                                                                          'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.__fetch_data': ( '00_fred/fred_api.html#fred.__fetch_data',
                                                                                    'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.__get_search_results': ( '00_fred/fred_api.html#fred.__get_search_results',
                                                                                            'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.__init__': ( '00_fred/fred_api.html#fred.__init__',
                                                                                'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred._parse': ( '00_fred/fred_api.html#fred._parse',
                                                                              'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.get_series': ( '00_fred/fred_api.html#fred.get_series',
                                                                                  'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.get_series_all_releases': ( '00_fred/fred_api.html#fred.get_series_all_releases',
                                                                                               'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.get_series_as_of_date': ( '00_fred/fred_api.html#fred.get_series_as_of_date',
                                                                                             'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.get_series_first_release': ( '00_fred/fred_api.html#fred.get_series_first_release',
                                                                                                'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.get_series_info': ( '00_fred/fred_api.html#fred.get_series_info',
                                                                                       'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.get_series_latest_release': ( '00_fred/fred_api.html#fred.get_series_latest_release',
                                                                                                 'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.get_series_vintage_dates': ( '00_fred/fred_api.html#fred.get_series_vintage_dates',
                                                                                                'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.search': ( '00_fred/fred_api.html#fred.search',
                                                                              'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.search_by_category': ( '00_fred/fred_api.html#fred.search_by_category',
                                                                                          'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.search_by_release': ( '00_fred/fred_api.html#fred.search_by_release',
                                                                                         'finsets/fred/fred_api.py')},
            'finsets.papers.hassan_etal_2019': { 'finsets.papers.hassan_etal_2019.clean': ( '02_papers/hassan_etal_2019.html#clean',
                                                                                            'finsets/papers/hassan_etal_2019.py'),
                                                 'finsets.papers.hassan_etal_2019.download': ( '02_papers/hassan_etal_2019.html#download',
                                                                                               'finsets/papers/hassan_etal_2019.py'),
                                                 'finsets.papers.hassan_etal_2019.source_url': ( '02_papers/hassan_etal_2019.html#source_url',
                                                                                                 'finsets/papers/hassan_etal_2019.py'),
                                                 'finsets.papers.hassan_etal_2019.variables': ( '02_papers/hassan_etal_2019.html#variables',
                                                                                                'finsets/papers/hassan_etal_2019.py')},
            'finsets.wrds.compa': { 'finsets.wrds.compa.book_equity': ('01_wrds/compa.html#book_equity', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.cashflow_vars': ('01_wrds/compa.html#cashflow_vars', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.clean': ('01_wrds/compa.html#clean', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.default_raw_vars': ('01_wrds/compa.html#default_raw_vars', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.download': ('01_wrds/compa.html#download', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.investment_vars': ('01_wrds/compa.html#investment_vars', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.issuance_vars': ('01_wrds/compa.html#issuance_vars', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.leverage_vars': ('01_wrds/compa.html#leverage_vars', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.liquidity_vars': ('01_wrds/compa.html#liquidity_vars', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.payout_vars': ('01_wrds/compa.html#payout_vars', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.profitability_vars': ( '01_wrds/compa.html#profitability_vars',
                                                                               'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.raw_metadata': ('01_wrds/compa.html#raw_metadata', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.raw_metadata_extra': ( '01_wrds/compa.html#raw_metadata_extra',
                                                                               'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.tobin_q': ('01_wrds/compa.html#tobin_q', 'finsets/wrds/compa.py')},
            'finsets.wrds.crspm': { 'finsets.wrds.crspm.clean': ('01_wrds/crspm.html#clean', 'finsets/wrds/crspm.py'),
                                    'finsets.wrds.crspm.default_raw_vars': ('01_wrds/crspm.html#default_raw_vars', 'finsets/wrds/crspm.py'),
                                    'finsets.wrds.crspm.delist_adj_ret': ('01_wrds/crspm.html#delist_adj_ret', 'finsets/wrds/crspm.py'),
                                    'finsets.wrds.crspm.download': ('01_wrds/crspm.html#download', 'finsets/wrds/crspm.py'),
                                    'finsets.wrds.crspm.parse_varlist': ('01_wrds/crspm.html#parse_varlist', 'finsets/wrds/crspm.py'),
                                    'finsets.wrds.crspm.raw_metadata': ('01_wrds/crspm.html#raw_metadata', 'finsets/wrds/crspm.py'),
                                    'finsets.wrds.crspm.raw_metadata_extra': ( '01_wrds/crspm.html#raw_metadata_extra',
                                                                               'finsets/wrds/crspm.py')},
            'finsets.wrds.linking': { 'finsets.wrds.linking.gvkey_permno_a': ( '01_wrds/linking.html#gvkey_permno_a',
                                                                               'finsets/wrds/linking.py'),
                                      'finsets.wrds.linking.gvkey_permno_m': ( '01_wrds/linking.html#gvkey_permno_m',
                                                                               'finsets/wrds/linking.py'),
                                      'finsets.wrds.linking.gvkey_permno_q': ( '01_wrds/linking.html#gvkey_permno_q',
                                                                               'finsets/wrds/linking.py')},
            'finsets.wrds.wrds_api': { 'finsets.wrds.wrds_api.Connection': ('01_wrds/wrds_api.html#connection', 'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__check_schema_perms': ( '01_wrds/wrds_api.html#connection.__check_schema_perms',
                                                                                                  'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__create_pgpass_file_unix': ( '01_wrds/wrds_api.html#connection.__create_pgpass_file_unix',
                                                                                                       'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__create_pgpass_file_win32': ( '01_wrds/wrds_api.html#connection.__create_pgpass_file_win32',
                                                                                                        'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__enter__': ( '01_wrds/wrds_api.html#connection.__enter__',
                                                                                       'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__exit__': ( '01_wrds/wrds_api.html#connection.__exit__',
                                                                                      'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__get_schema_for_view': ( '01_wrds/wrds_api.html#connection.__get_schema_for_view',
                                                                                                   'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__get_user_credentials': ( '01_wrds/wrds_api.html#connection.__get_user_credentials',
                                                                                                    'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__init__': ( '01_wrds/wrds_api.html#connection.__init__',
                                                                                      'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__make_sa_engine_conn': ( '01_wrds/wrds_api.html#connection.__make_sa_engine_conn',
                                                                                                   'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__write_pgpass_file': ( '01_wrds/wrds_api.html#connection.__write_pgpass_file',
                                                                                                 'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.close': ( '01_wrds/wrds_api.html#connection.close',
                                                                                   'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.connect': ( '01_wrds/wrds_api.html#connection.connect',
                                                                                     'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.create_pgpass_file': ( '01_wrds/wrds_api.html#connection.create_pgpass_file',
                                                                                                'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.describe_table': ( '01_wrds/wrds_api.html#connection.describe_table',
                                                                                            'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.get_row_count': ( '01_wrds/wrds_api.html#connection.get_row_count',
                                                                                           'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.get_table': ( '01_wrds/wrds_api.html#connection.get_table',
                                                                                       'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.list_libraries': ( '01_wrds/wrds_api.html#connection.list_libraries',
                                                                                            'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.list_tables': ( '01_wrds/wrds_api.html#connection.list_tables',
                                                                                         'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.load_library_list': ( '01_wrds/wrds_api.html#connection.load_library_list',
                                                                                               'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.raw_sql': ( '01_wrds/wrds_api.html#connection.raw_sql',
                                                                                     'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.NotSubscribedError': ( '01_wrds/wrds_api.html#notsubscribederror',
                                                                                     'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.SchemaNotFoundError': ( '01_wrds/wrds_api.html#schemanotfounderror',
                                                                                      'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.download': ('01_wrds/wrds_api.html#download', 'finsets/wrds/wrds_api.py')}}}
